package sk.upjs.paz.heap;

import java.awt.Color;
import java.awt.Font;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;

import sk.upjs.jpaz2.*;

public class HeapPane extends Pane {

    /**
     * Internal class for animating
     */
    private class PaneSwapAnimation extends TickTimer {
	static final int SWAP_DURATION = 1000;
	Pane pane1;
	Pane pane2;
	Point2D pane1Start;
	Point2D pane2Start;
	long startTime;

	public PaneSwapAnimation(Pane pane1, Pane pane2) {
	    this.pane1 = pane1;
	    this.pane2 = pane2;
	    pane1Start = pane1.getPosition();
	    pane2Start = pane2.getPosition();

	    startTime = System.currentTimeMillis();
	    setTickPeriod(50);
	    setEnabled(true);
	}

	/**
	 * Stops animation of swaping
	 */
	public void stop() {
	    setEnabled(false);
	    pane1.setPosition(pane2Start);
	    pane2.setPosition(pane1Start);
	    swapAnimation = null;
	}

	@Override
	protected void onTick() {
	    double progress = (System.currentTimeMillis() - startTime) / (double) SWAP_DURATION;
	    if (progress >= 1) {
		stop();
		return;
	    }

	    pane1.setPosition(pane1Start.getX() + (pane2Start.getX() - pane1Start.getX()) * progress, pane1Start.getY()
		    + (pane2Start.getY() - pane1Start.getY()) * progress);
	    pane2.setPosition(pane2Start.getX() + (pane1Start.getX() - pane2Start.getX()) * progress, pane2Start.getY()
		    + (pane1Start.getY() - pane2Start.getY()) * progress);
	}
    }

    /**
     * Positions of field centers
     */
    private Point2D[] fieldCenters;

    /**
     * Values at fields
     */
    private NumberValue[] values;

    /**
     * Length of active subarray
     */
    private int activeLength;

    /**
     * Show indices of fields
     */
    private boolean displayedIndices;
    
    /**
     * Show comparators
     */
    private boolean displayComparators;
    
    /**
     * Swap animation in progress
     */
    private PaneSwapAnimation swapAnimation = null;

    /**
     * Constructs the pane
     */
    public HeapPane() {
	super(900, 520);
	setBorderWidth(0);
	initiate(15);
	drawBackground();
    }

    /**
     * Sets the size of the heap (the number of active fields)
     */
    public void setSize(int newSize) {
	if (newSize < 0)
	    newSize = 0;

	if (newSize > 16)
	    newSize = 16;

	stopAnimation();
	activeLength = newSize;
	drawBackground();
    }

    /**
     * Returns the size of the heap
     */
    public int getSize() {
	return activeLength;
    }

    /**
     * Returns the maximum possible size
     */
    public int getMaxSize() {
	return values.length;
    }
    
    /**
     * Sets values of fields
     */
    public void setValues(int[] numbers) {
	stopAnimation();
	for (int i = 0; i < Math.min(values.length, numbers.length); i++) {
	    values[i].setValue(numbers[i]);
	}
	
	if (displayComparators) {
	    drawBackground();
	}
    }
    
    public boolean isDisplayedIndices() {
        return displayedIndices;
    }

    public void setDisplayedIndices(boolean displayedIndices) {
        stopAnimation();
        
	this.displayedIndices = displayedIndices;
        drawBackground();
    }

    public boolean isDisplayComparators() {
        return displayComparators;
    }

    public void setDisplayComparators(boolean displayComparators) {
	stopAnimation();
	
	this.displayComparators = displayComparators;
        drawBackground();
    }

    
    /**
     * Swaps the first and the last element
     */
    public void swapFirstAndLast() {
	if (activeLength < 2)
	    return;
	
	animateSwap(0, activeLength-1);
    }

    /**
     * Computes field centers for all ancestor nodes.
     * 
     * @param idx
     *            index of the root element
     * @param gapLevel
     *            distance between two consecutive levels
     * @param width
     *            with of the tree rooted at this index
     */
    private void computeFieldCentersOfChildren(int idx, double gapLevel, double width) {
	int leftIdx = 2 * idx + 1;
	int rightIdx = 2 * idx + 2;

	if (leftIdx < fieldCenters.length) {
	    fieldCenters[leftIdx] = new Point2D.Double(fieldCenters[idx].getX() - width / 4, fieldCenters[idx].getY()
		    + gapLevel);
	    computeFieldCentersOfChildren(leftIdx, gapLevel, width / 2);
	}

	if (rightIdx < fieldCenters.length) {
	    fieldCenters[rightIdx] = new Point2D.Double(fieldCenters[idx].getX() + width / 4, fieldCenters[idx].getY()
		    + gapLevel);
	    computeFieldCentersOfChildren(rightIdx, gapLevel, width / 2);
	}
    }

    /**
     * Computes field centers and creates object visualizing field values.
     */
    private void initiate(int length) {
	fieldCenters = new Point2D[length];
	fieldCenters[0] = new Point2D.Double(getWidth() / 2, 60);
	computeFieldCentersOfChildren(0, 130, getWidth());

	values = new NumberValue[length];
	for (int i = 0; i < length; i++) {
	    values[i] = new NumberValue();
	    values[i].setPosition(fieldCenters[i]);
	    add(values[i]);
	}

	activeLength = length;
    }

    /**
     * Draws the background
     */
    private void drawBackground() {
	clear();

	Turtle drawer = new Turtle();
	drawer.setVisible(false);
	add(drawer);

	// Draw arcs
	drawer.setPenWidth(5);
	for (int i = 1; i < values.length; i++) {
	    int parentIdx = (i - 1) / 2;
	    boolean isActive = (i < activeLength);
	    drawer.setPenColor(isActive ? Color.BLACK : Color.LIGHT_GRAY);
	    drawer.setPosition(fieldCenters[i]);
	    drawer.moveTo(fieldCenters[parentIdx]);
	    if (displayComparators && isActive) {
		drawer.setFont(drawer.getFont().deriveFont(14f).deriveFont(Font.BOLD));
		drawer.penUp();
		drawer.setPosition((fieldCenters[i].getX() + fieldCenters[parentIdx].getX())/2, (fieldCenters[i].getY() + fieldCenters[parentIdx].getY())/2);
		
		if (values[parentIdx].getValue() >= values[i].getValue()) {
		    drawer.setFillColor(Color.green);
		} else {
		    drawer.setFillColor(Color.red);
		}
		
		drawer.dot(15);
		drawer.turnTowards(fieldCenters[i]);
		drawer.printCenter("≥");
		drawer.penDown();
	    }
	}

	drawer.penUp();
	drawer.setPenColor(Color.blue);
	drawer.setFont(drawer.getFont().deriveFont(18f));
	
	// Draw fields
	for (int i = 0; i < values.length; i++) {
	    drawer.setFillColor((i < activeLength) ? Color.BLACK : Color.LIGHT_GRAY);
	    drawer.setPosition(fieldCenters[i]);
	    drawer.dot(35);
	    if (displayedIndices) {
		drawer.setDirection(0);
		if ((i-1) % 2 == 0)
		    drawer.turn(-45);
		else
		    drawer.turn(45);
		
		drawer.step(50);
		drawer.setDirection(90);
		drawer.printCenter(Integer.toString(i));
	    }
	}

	remove(drawer);
    }

    /**
     * Computes distance between two points.
     */
    private double distanceBetween(Point2D point1, Point2D point2) {
	double dx = point1.getX() - point2.getX();
	double dy = point1.getY() - point2.getY();
	return Math.sqrt(dx * dx + dy * dy);
    }

    /**
     * Returns a field index such that a point [x, y] lies on the edge from
     * indexed field to its parent.
     */
    private int arcFromIndex(double x, double y) {
	Point2D point = new Point2D.Double(x, y);
	for (int i = 1; i < values.length; i++) {
	    int parentIdx = (i - 1) / 2;
	    double pointToI = distanceBetween(point, fieldCenters[i]);
	    double pointToParent = distanceBetween(point, fieldCenters[parentIdx]);
	    double IToParent = distanceBetween(fieldCenters[i], fieldCenters[parentIdx]);

	    if ((pointToI >= 35) && (pointToParent >= 35) && (pointToI + pointToParent <= IToParent + 3))
		return i;
	}

	return -1;
    }

    /**
     * Stops a swap animation if there is such an animation in progress.
     */
    private void stopAnimation() {
	if (swapAnimation != null) {
	    swapAnimation.stop();
	    swapAnimation = null;
	}
    }
    
    /**
     * Swaps the values and starts animation of this swap
     */
    private void animateSwap(int idx1, int idx2) {
	if (idx1 == idx2)
	    return;
	
	stopAnimation();
	swapAnimation = new PaneSwapAnimation(values[idx1], values[idx2]);
	NumberValue tmp = values[idx1];
	values[idx1] = values[idx2];
	values[idx2] = tmp;
	
	if (displayComparators)
	    drawBackground();
    }

    @Override
    protected boolean onCanClick(int x, int y) {
	if (swapAnimation != null)
	    return false;

	int arcIdx = arcFromIndex(x, y);
	return (arcIdx != -1) && (arcIdx < activeLength);
    }

    @Override
    protected void onMousePressed(int x, int y, MouseEvent detail) {
	if (swapAnimation != null)
	    return;

	int arcIdx = arcFromIndex(x, y);
	if (!((arcIdx > 0) && (arcIdx < activeLength)))
	    return;

	animateSwap(arcIdx, (arcIdx - 1) / 2);
    }
}
